@@ -1,6 +1,6 @@
 from __future__ import annotations
 import threading
-
+from copy import deepcopy
 import json
 from typing import List, Dict, Any
 import os
@@ -22,6 +22,9 @@ from magicdb.utils.Serverless.span import safe_span
 
 from magicdb.utils.updating_objects import make_update_obj
 
+from magicdb.utils.async_helpers import threadpool_asyncio, threadpool, magic_async
+
+
 from slugify import slugify as base_slugify
 
 slugify_kwargs = {"separator": "", "replacements": [["&", "and"]]}
@@ -77,21 +80,22 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
 
     def __init__(self, from_db: bool = False, **kwargs):
         """Feed in all magic fields as kwargs"""
-        kwargs_copy = kwargs.copy()
+        # TODO test how slow this deep copy is
+        kwargs_from_db = {} if not from_db else deepcopy(kwargs)
         model_helpers.make_magic_fields_from_kwargs(
             kwargs=kwargs, collection_name=self.collection_name
         )
         super().__init__(**kwargs)
-        self.kwargs_from_db = {} if not from_db else kwargs_copy
+        self.kwargs_from_db = kwargs_from_db
 
     @classmethod
     def construct(cls, *args, from_db: bool = False, **kwargs):
-        kwargs_copy = kwargs.copy()
+        kwargs_from_db = {} if not from_db else deepcopy(kwargs)
         model_helpers.make_magic_fields_from_kwargs(
             kwargs=kwargs, collection_name=cls.get_collection_name()
         )
         new_obj = super().construct(*args, **kwargs)
-        new_obj.kwargs_from_db = {} if not from_db else kwargs_copy
+        new_obj.kwargs_from_db = kwargs_from_db
         return new_obj
 
     """GETTING AND SETTING FIELDS"""
@@ -130,7 +134,7 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
         given_fields_to_exclude = getattr(cls.Meta, "exclude", set())
         return magic_fields_to_exclude | given_fields_to_exclude
 
-    def dict(self, *args, exclude_magic_fields=True, **kwargs):
+    def dict(self, exclude_magic_fields=True, *args, **kwargs):
         to_exclude = set() if not exclude_magic_fields else self.get_fields_to_exclude()
         # join w other fields passed in
         kwargs["exclude"] = kwargs.get("exclude") or set() | to_exclude
@@ -273,7 +277,15 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
         self.__dict__.update(obj_to_save.__dict__)
         return obj_to_save
 
-    def update(self, batch=None, transaction=None, create=False, ignore_fields=False):
+    def update(
+        self,
+        batch=None,
+        transaction=None,
+        create=False,
+        ignore_fields=False,
+        print_update_d=False,
+        only_print_update_d=False,
+    ):
         batch_or_transaction = batch or transaction
         # TODO this takes a long time... find out why and fix! Do you need to make new class?
         obj_to_update = (
@@ -282,23 +294,31 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
             else self.__class__(**self.dict(exclude_magic_fields=False))
         )
         new_d = obj_to_update.dict()
+        kwargs_d = self.kwargs_from_db
 
         self.remove_magic_fields(new_d)
+        self.remove_magic_fields(kwargs_d)
 
         self.get_fields_to_exclude()
         update_d = (
-            new_d
-            if not obj_to_update.kwargs_from_db
-            else make_update_obj(original=obj_to_update.kwargs_from_db, new=new_d)
+            new_d if not kwargs_d else make_update_obj(original=kwargs_d, new=new_d)
         )
-
+        if print_update_d or only_print_update_d:
+            pprint.pprint({f"update_d_for_{self.id}": update_d})
+        if only_print_update_d:
+            return
         try:
-            with safe_span(f"update-{self.key}", use=(batch_or_transaction is None)):
-                obj_to_update.ref.update(
-                    update_d
-                ) if not batch_or_transaction else batch_or_transaction.update(
-                    obj_to_update.ref, update_d
-                )
+            if update_d != {}:
+                with safe_span(
+                    f"update-{self.key}", use=(batch_or_transaction is None)
+                ):
+                    obj_to_update.ref.update(
+                        update_d
+                    ) if not batch_or_transaction else batch_or_transaction.update(
+                        obj_to_update.ref, update_d
+                    )
+            else:
+                print(f"update_d was empty so did not update for {self.id}.")
             self.__dict__.update(obj_to_update.__dict__)
             return obj_to_update
         except Exception as e:
@@ -308,6 +328,7 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
                 else:
                     db_error = DatabaseError(obj_to_update.key)
                     raise DatabaseError(db_error.message)
+            raise e
 
     def delete(self, batch=None, transaction=None):
         batch_or_transaction = batch or transaction
@@ -320,7 +341,7 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
 
     """Save history to DB"""
 
-    def save_history(self, sync: bool = True, user_id: str = None):
+    def save_history(self, sync: bool = True, user_id: str = None, **kwargs):
         """Saves current object to a changes sub collection"""
         history_key = f"{self.key}/history/{str(int(time.time()))}"
 
@@ -329,9 +350,9 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
             changed_by: str = user_id
 
         history_obj = History(**self.dict(), key=history_key)
-        return history_obj.save() if sync else history_obj.save_async()
+        return history_obj.save(**kwargs) if sync else history_obj.save_async(**kwargs)
 
-    def save_deleted(self, sync: bool = True, user_id: str = None):
+    def save_deleted(self, sync: bool = True, user_id: str = None, **kwargs):
         deleted_key = f"deleted_{self.collection_name}/{self.id}"
 
         class Deleted(self.__class__):
@@ -339,7 +360,7 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
             deleted_by: str = user_id
 
         deleted_obj = Deleted(**self.dict(), key=deleted_key)
-        return deleted_obj.save() if sync else deleted_obj.save_async()
+        return deleted_obj.save(**kwargs) if sync else deleted_obj.save_async(**kwargs)
 
     """Writing to DB async"""
 
@@ -376,6 +397,7 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
         return getattr(cls.Meta, "default_exception", Exception)
 
     @classmethod
+    @magic_async
     def get_with_error(cls, id, error=None, **kwargs):
         if not error:
             error = cls.get_default_exception()
@@ -384,6 +406,31 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
             raise error(f"{cls.__name__} with id {id} does not exist.")
         return val
 
+    @classmethod
+    @threadpool_asyncio
+    def get_with_error_async(cls, id, error=None, **kwargs):
+        if not error:
+            error = cls.get_default_exception()
+        val = cls.collection.get(id, **kwargs)
+        if not val:
+            raise error(f"{cls.__name__} with id {id} does not exist.")
+        return val
+
+    @classmethod
+    @threadpool_asyncio
+    def get_async(cls, id, **kwargs):
+        return cls.collection.get(id, **kwargs)
+
+    @classmethod
+    @threadpool
+    def get_with_error_future(cls, id, error=None, **kwargs):
+        if not error:
+            error = cls.get_default_exception()
+        val = cls.collection.get(id, **kwargs)
+        if not val:
+            raise error(f"{cls.__name__} with id {id} does not exist.")
+        return val
+
     """GETTING SUBCLASSES"""
 
     @classmethod
@@ -404,9 +451,9 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
 
     @staticmethod
     def stream_queries(queries: List[magicdb.Query]):
-        with concurrent.futures.ThreadPoolExecutor() as executor:
+        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
             # set this to 10 because lambda only has 2 cores so would only have 6 threads for max_workers
-            executor._max_workers = max(executor._max_workers, 20)
+            # executor._max_workers = max(executor._max_workers, 20)
             futures = [
                 executor.submit(
                     query.stream,
@@ -454,5 +501,43 @@ class MagicModel(BaseModel, metaclass=QueryAndBaseMetaClasses):
         models = cls.collection.where("slug", "==", slug).stream()
         return None if not models else models[0]
 
+    @classmethod
+    def get_by_slug_with_error(cls, slug, error=None):
+        if not error:
+            error = cls.get_default_exception()
+        models = cls.collection.where("slug", "==", slug).stream()
+        if not models:
+            raise error(f"{cls.__name__} with slug {slug} does not exist.")
+        return models[0]
+
+    @classmethod
+    @magic_async
+    def get_by_field(cls, field_name: str, value):
+        models = cls.collection.where(
+            magicdb.db.conn.field_path(field_name), "==", value
+        ).stream()
+        return None if not models else models[0]
+
+    @classmethod
+    @threadpool_asyncio
+    def get_by_field_async(cls, field_name: str, value):
+        return cls.get_by_field(field_name=field_name, value=value)
+
+    @classmethod
+    def get_by_field_with_error(cls, field_name: str, value, error=None):
+        if not error:
+            error = cls.get_default_exception()
+        model = cls.get_by_field(field_name=field_name, value=value)
+        if not model:
+            raise error(f"{cls.__name__} with {field_name} {value} does not exist.")
+        return model
+
+    @classmethod
+    @threadpool_asyncio
+    def get_by_field_with_error_async(cls, field_name: str, value, error=None):
+        return cls.get_by_field_with_error(
+            field_name=field_name, value=value, error=error
+        )
+
 
 MagicModel.update_forward_refs()
