@@ -1,6 +1,7 @@
 from typing import Union, List
 from magicdb import db, ASCENDING, DESCENDING, DocumentSnapshot
 from magicdb.utils.Serverless.span import safe_span
+from magicdb.utils.async_helpers import threadpool_asyncio, threadpool, magic_async
 
 import time
 
@@ -64,6 +65,14 @@ class Query:
         self.firebase_query = self.firebase_query.start_at(fields)
         return self
 
+    def start_after_id(self, last_doc_id: str = None):
+        if not last_doc_id:
+            return self
+        key = self.cls.construct(id=last_doc_id).key
+        snapshot = db.conn.document(key).get()
+        self.firebase_query = self.firebase_query.start_after(snapshot)
+        return self
+
     def start_after(self, fields):
         self.firebase_query = self.firebase_query.start_after(fields)
         return self
@@ -78,7 +87,6 @@ class Query:
 
     def limit(self, limit):
         if not limit or limit == -1:
-            print("no limit given")
             return self
         self.firebase_query = self.firebase_query.limit(limit)
         return self
@@ -96,6 +104,7 @@ class Query:
     def make_safe_span_label(self, query_type):
         return f"{query_type}-{self.get_path_of_query()}-{str(self.query_inputs)}"
 
+    @magic_async
     def stream(self, validate_db_data=True, **kwargs):
         """Currently defaults to validating the db data, but you can turn this off with the validate_db_data param"""
         with safe_span(self.make_safe_span_label("stream")):
@@ -108,6 +117,19 @@ class Query:
             for doc in docs
         ]
 
+    @threadpool_asyncio
+    def stream_async(self, validate_db_data=True, **kwargs):
+        with safe_span(self.make_safe_span_label("stream")):
+            docs = list(self.firebase_query.stream(**kwargs))
+
+        constructor = self.cls if validate_db_data else self.cls.construct
+
+        return [
+            constructor(from_db=True, doc=doc, **doc.to_dict(), key=doc.reference.path)
+            for doc in docs
+        ]
+
+    @magic_async
     def get(self, id=None, create=False, validate_db_data=True, **kwargs):
         if id:
             self.firebase_query = self.firebase_query.document(id)
@@ -131,14 +153,25 @@ class Query:
             key=doc.reference.path,
         )
 
+    @staticmethod
+    def sort_models_by_id_order(models, ids):
+        """Maintain order of return array"""
+        sorted_models = [None] * len(ids)
+        sort_key = {id: index for index, id in enumerate(ids)}
+        for model in models:
+            sorted_models[sort_key[model.id]] = model
+        return sorted_models
+
+    @magic_async
     def get_all(self, ids, validate_db_data=True, **kwargs):
         doc_refs = [db.conn.collection(self.collection_path).document(id) for id in ids]
 
         with safe_span(self.make_safe_span_label("get_all")):
-            docs = db.conn.get_all(doc_refs, **kwargs)
+            docs = list(db.conn.get_all(doc_refs, **kwargs))
 
         constructor = self.cls if validate_db_data else self.cls.construct
-        return [
+
+        models = [
             constructor(
                 from_db=True,
                 doc=doc,
@@ -147,6 +180,11 @@ class Query:
             )
             for doc in docs
         ]
+        return self.sort_models_by_id_order(models, ids)
+
+    @threadpool_asyncio
+    def get_all_async(self, ids, validate_db_data=True, **kwargs):
+        return self.get_all(ids, validate_db_data, **kwargs)
 
     """Subcollections"""
 
